<?php

/**
 * Mail_DSN - process e-mail bounces/Delivery Status Notifications (DSNs)
 *
 * (c)2004-2008 Tim Jackson (tim@timj.co.uk)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of version 3 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @package Mail_DSN
 */

// TODO: use mimeDecode to decode MIME parts

require_once 'Mail/DSN/Metadata.php';

class Mail_DSN {

	// Type of bounce
	const TYPE_TEMPORARY  = 'temp';
	const TYPE_PERMANENT  = 'perm';
	const TYPE_AUTOREPLY  = 'autoreply';
	const TYPE_CHALLENGE  = 'challenge'; // Some kind of C-R challenge
	const TYPE_QUARANTINE = 'quarantine'; // Like temp, but where sender needs to do something
	const TYPE_READ       = 'read'; // Notifications that the message has been read
	const TYPE_UNKNOWN    = 'unknown';

	// Reasons for DSN
	const REASON_SMTP_MAIL = 'smtp_mail';
	const REASON_SMTP_RCPT = 'smtp_rcpt';
	const REASON_SMTP_DATA = 'smtp_data';
	const REASON_RECIPIENT_UNKNOWN = 'recipient_unknown';
	const REASON_RECIPIENT_DISABLED = 'disabled_recipient';
	const REASON_RECIPIENT_SYNTAX = 'recipient_syntax';
	const REASON_RECIPIENT_BLOCKED_SENDER = 'recipient_blocked_sender';
	const REASON_RECIPIENT_DNS = 'dns';
	const REASON_NO_RECIPIENTS = 'no_recipients';
	const REASON_SENDER_DNS = 'sender_dns';
	const REASON_BLACKLISTED_HOST = 'host_blacklisted';
	const REASON_BLACKLISTED_DOMAIN = 'domain_blacklisted';
	const REASON_BLACKLISTED = 'blacklisted';
	const REASON_CONNECTION_FAILURE = 'connection_failure';
	const REASON_DELAY = 'delay'; // Generic delay
	const REASON_UNKNOWN = 'unknown';
	const REASON_OVERQUOTA = 'overquota';
	const REASON_CONTENT = 'content'; // Something to do with the content of the mail
	const REASON_RELAY_DENIED = 'relay_denied'; // Almost certainly indicates a problem at the remote end
	const REASON_CANCELLED = 'cancelled'; // cancelled by admin etc.
	

	// MTAs	
	const MTA_UNKNOWN    = 'unknown';
	const MTA_SENDMAIL   = 'sendmail';
	const MTA_EXIM       = 'exim';
	const MTA_POSTFIX    = 'postfix';
	const MTA_QMAIL      = 'qmail';
	const MTA_EXCHANGE   = 'exchange';
	const MTA_INTERMAIL  = 'intermail';
	const MTA_MERAK      = 'merak';
	const MTA_SIMS       = 'sims';// Sun Internet Mail Server
	const MTA_PROCESS    = 'pmdf'; //Process Software (PMDF)
	const MTA_GROUPWISE  = 'groupwise';
	const MTA_LOTUSNOTES = 'lotusnotes';
	
	/**
	 * Common IP-based blacklists
	 */
	protected $_blacklists_ip = array('spamhaus.org','spews.org','blackholes.us','dnsbl.ahbl.org','dsbl.org','five-ten-sg.com','cbl.abuseat.org');

	/**
	 * Common domain-based blacklists
	 */
	protected $_blacklists_domain = array('rhsbl.ahbl.org','dsn.rfc-ignorant.org','postmaster.rfc-ignorant.org','abuse.rfc-ignorant.org','bogusmx.rfc-ignorant.org','whois.rfc-ignorant.org');
	
	/**
	 * Header that can optionally be used on outgoing mail to help us track 
	 * bounces that come back in
	 */
	protected static $_tracking_header = 'X-DSN-Tracker';

	/**
	 * Process a DSN to extract information from it
	 *
	 * @param string $message  The complete raw DSN message
	 * @return Mail_DSN_Metadata  Metadata about the DSN
	 */	
	/**
	 * Process a DSN message
	 *
	 * @param string $message     Raw e-mail source of DSN
	 * @return Mail_DSN_Metadata  Metadata about the bounce
	 */
	public function processMessage($message)
	{
		$this->checkRuntimeDependencies();
	
		if (empty($message)) {
			throw new Exception('Mail_DSN::processMessage() called on empty message!');
		}
		
		$dsn = new Mail_DSN_Metadata;
		//$MailDecoded = Mail_mimeDecode::decode($MIMEOptions);
		// Exim
		if (preg_match('/This message was created automatically by mail delivery software( \(Exim\))?./', $message) || strstr($message,'The following text was generated during the delivery attempt(s):') || strstr($message,'------ This is a copy of your message, including all the headers. ------') || preg_match('/^Subject: Mail delivery failed: returning message to sender/m',$message)) {
			$dsn->mta = self::MTA_EXIM;
			
			if (strstr($message,'This is a permanent error.')) {
				$dsn->type = self::TYPE_PERMANENT;
			} else if (strstr($message, 'Delay reason:')) {
				$dsn->type = self::TYPE_TEMPORARY;
			} else if (strstr($message, 'delivery cancelled by administrator')) {
				$dsn->type = self::TYPE_CANCELLED;
			}
			
			if (preg_match('/The following address\(es\) failed:\s*(\S+)(\n.+\n)?/', $message, $matches)) {
				if (strstr($matches[1],'@')) {
					$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
					if (isset($matches[2]) && !empty($matches[2])) {
						$dsn->message = trim($matches[2]);
					}
				}
			} else if (preg_match('/The address to which the message has not yet been delivered is:\n\n(.+)\n(.+)\n/', $message, $matches)) {
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
				$dsn->message = $this->_cleanString($matches[2]);
			} else if (preg_match('/^A message that you sent contained one or more recipient addresses that were\s+incorrectly constructed:\n\n(.+):(.+)\n/Um',$message,$matches)) {
				$dsn->type = self::TYPE_PERMANENT;
				$dsn->reason = self::REASON_RECIPIENT_SYNTAX;
				$dsn->original_recipient = trim($matches[1]);// DO NOT canonicalise!
				$dsn->message = $this->_cleanString($matches[2]);
			} else if (preg_match('/^A message that you sent contained no recipient addresses, and therefore no\s+delivery could be attempted./m',$message,$matches)) {
				$dsn->type = self::TYPE_PERMANENT;
				$dsn->reason = self::REASON_NO_RECIPIENTS;
				$dsn->message = $this->_cleanString($matches[0]);
			}
			
			// Fallback for weird internal errors
			if ($dsn->type == self::TYPE_UNKNOWN && preg_match('/^The following text was generated during the delivery attempt\(s\):\n\n\s*(.+)\n(.+)\n\n/Usm',$message,$matches)) {
				$dsn->type = self::TYPE_PERMANENT; // a bit of an assumption
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
				if (isset($matches[2]) && !empty($matches[2])) {
					$dsn->message = $this->_cleanString($matches[2]);
				}
			}
			
			// Remote rejection
			if (preg_match('/SMTP error from remote mailer after ([^:]+):/',$message,$matches)) {
				$dsn->message = null; // clear message fragment that might already be set
				switch ($matches[1]) {
					case 'MAIL FROM':
						$dsn->reason = self::REASON_SMTP_MAIL;
						break;
					case 'RCPT TO':
						$dsn->reason = self::REASON_SMTP_RCPT;
						break;
					case 'end of data':
						$dsn->reason = self::REASON_SMTP_DATA;
						break;
				}
				if (preg_match('/host \S+ \[\S+\]:\n?(.+)\n\n/Uis',$message,$matches)) {
					$dsn->reason = $this->_processReason($matches[1],$dsn->reason);
					$this->_processSMTPError($dsn,$matches[1]);
				}
				
			} else {
				// Locally-generated bounce (not necessarily local to the current machine)
			}
			
			if ($dsn->reason == self::REASON_UNKNOWN && strstr($dsn->message,'host lookup did not complete')) {
				$dsn->reason = self::REASON_RECIPIENT_DNS;
			}
		}
		
		// Qmail
		if (strstr($message,'This is the qmail-send program') || preg_match('/^Subject: failure (notice|delivery)$/m',$message)) {
			$dsn->mta = self::MTA_QMAIL;
			if (strstr($message,'This is a permanent error') || strstr($message,'Unable to deliver message to the following address')) {
				$dsn->type = self::TYPE_PERMANENT;
			}
			
			if (preg_match('/(Sorry it didn\'t work out|to the following address\(es\))\.\n\n<(.+)>:\n(.+)\n(.*)\n/Ums',$message,$matches)) {
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[2]);
				$dsn->message = $this->_cleanString($matches[3]);
				if (isset($matches[4]) && preg_match('/Remote host said: (.+)/', $matches[4], $submatches)) {
					$dsn->message = null;
					$this->_processSMTPError($dsn,$submatches[1]);
				}
			}
		}
		
		// Sendmail
		if (preg_match('/The following addresses had (permanent |transient non-)fatal errors/', $message)|| preg_match('/^Received-from-MTA:/im',$message)) {
			$dsn->mta = self::MTA_SENDMAIL;
			
			if (preg_match('/The following addresses had (permanent |transient non-)fatal errors/', $message, $matches)) {
				if (strstr($matches[1],'permanent')) {
					$dsn->type = self::TYPE_PERMANENT;
				} else {
					$dsn->type = self::TYPE_TEMPORARY;
				}
			}
			
			if (preg_match('/(RCPT To:|Final-Recipient: (RFC|rfc)822; )(.+)/',$message,$matches)) {
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[3]);
				if (strstr($matches[0],'RCPT To:') && $dsn->reason == self::REASON_UNKNOWN) {
					$dsn->reason = self::REASON_SMTP_RCPT;
				}
			}
			
			if (preg_match('/^(SMTP diagnostic:|Diagnostic-Code: SMTP;|<<<) (.+)/im',$message,$matches)) {
				$this->_processSMTPError($dsn,$matches[2]);
			}
			
			if ($dsn->reason == self::REASON_UNKNOWN && preg_match('/----- Transcript of session follows -----\n(.+)\n(.+\n)?/',$message,$matches)) {
				// Matches[1] might contain:
				// "...while talking to..."
				// <a message>
				// <an SMTP error>
				
				$SMTPError = null;
				$message = null;
				
				for ($x=1; $x<=2; $x++) {
					if (preg_match('/^\d\d\d /',$matches[$x])) {
						if (!$SMTPError) $SMTPError = $matches[$x];
					} else if (!strstr('while talking to',$matches[$x])) {
						if (!$message) $message = $matches[$x];
					}
				}
				
				if (empty($dsn->message) && $message) $dsn->message = $this->_cleanString($message);
				if ($SMTPError) $this->_processSMTPError($dsn,$SMTPError);
			}
		}
		
		// Postfix
		// also Subject: Undelivered Mail Returned to Sender
		if (strstr($message,'This is the Postfix program')) {
			$dsn->mta = self::MTA_POSTFIX;
			
			if (strstr($message,'could not be delivered to one or more destinations') || strstr($message,'Subject: Undelivered Mail Returned to Sender')) {
				$dsn->type = self::TYPE_PERMANENT;
			}
			
			if (preg_match('/The Postfix program\n\n<(.+)>: (.+)\n\n/Us',$message,$matches)) {
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
				$matches[2] = $this->_cleanString($matches[2]);
				if (preg_match('/^host \S+\[\S+\] said: (.+) \(in reply to (.+) command\)/',$matches[2],$submatches)) {
					switch ($submatches[2]) {
						case 'MAIL FROM':
							$dsn->reason = self::REASON_SMTP_MAIL;
							break;
						case 'RCPT TO':
							$dsn->reason = self::REASON_SMTP_RCPT;
							break;
						case 'end of DATA':
							$dsn->reason = self::REASON_SMTP_DATA;
							break;
					}
					$this->_processSMTPError($dsn,$submatches[1]);
				} else {
					$dsn->message = trim($matches[2]);
				}
			}
		}
		
		// Lotus notes
		if (preg_match('/^Subject: (Out of Office Notification:|.+(is out of the office|est absent\(e\))\.)/m',$message,$matches)) {
			$dsn->mta = self::MTA_LOTUSNOTES;
			$dsn->type = self::TYPE_AUTOREPLY;
			$dsn->message = $matches[1];
		}
		
		// Exchange
		if (strstr($message,'This is an automatically generated Delivery Status Notification.') || strstr($message,'Subject: Out of Office AutoReply:') || strstr($message,'X-Mailer: Internet Mail Service') || preg_match('/^Subject: Undeliverable:/m',$message)) {
			$dsn->mta = self::MTA_EXCHANGE;
			
			if (strstr($message,'THIS IS A WARNING MESSAGE ONLY.')) {
				$dsn->type = self::TYPE_TEMPORARY;
			} else if (strstr($message,'Subject: Out of Office AutoReply:')) {
				$dsn->type = self::TYPE_AUTOREPLY;
				// TODO: set $dsn->message to contain message body?
			}
			
			if (preg_match('/^Diagnostic-Code: SMTP;\s*(.+)/im',$message,$matches)) {
				$this->_processSMTPError($dsn,$matches[1]);
			}
			
			// is this hotmail-specific?
			if (preg_match('/Unable to deliver message to the following recipients, due to being unable to connect successfully to the destination mail server.\s*(\S+)/',$message,$matches)) {
				$dsn->type = self::TYPE_PERMANENT;
				$dsn->reason = self::REASON_CONNECTION_FAILURE;
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
				$dsn->message = 'Unable to deliver message to the following recipients, due to being unable to connect successfully to the destination mail server.';
			}
			
			// is this hotmail-specific?
			else if (preg_match('/Delivery to the following recipients has been delayed\.\n\n(.+)\n/',$message,$matches)) {
				$dsn->type = self::TYPE_TEMPORARY;
				$dsn->reason = self::REASON_DELAY;
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
				if (!$dsn->message) $dsn->message = 'Delivery to the following recipients has been delayed.';
			}
			
			// is this hotmail-specific?
			else if (preg_match('/Delivery to the following recipients failed\.\n\n(.+)\n/',$message,$matches)) {
				$dsn->type = self::TYPE_PERMANENT;
				if (!$dsn->message) $dsn->message = 'Delivery to the following recipients failed.';
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
			}
			
			else if (strstr($message,'Subject: Undeliverable:')) {
				$dsn->type = self::TYPE_PERMANENT;
				if (preg_match('/did not reach the following recipient\(s\):\n\n(\S+) on .+\n(.+)\n/',$message,$matches)) {
					$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
					$dsn->message = trim($matches[2]);
				}
			}
		}
		
		// Intermail
		if (preg_match('/^This Message was undeliverable due to the following reason:\n\n(.+)\n\n<(.+)>\n/m',$message,$matches)) {
			$dsn->mta = self::MTA_INTERMAIL;
			$dsn->type = self::TYPE_PERMANENT;
			$dsn->reason = $this->_processReason($matches[1],$dsn->reason);
			$dsn->message = $this->_cleanString($matches[1]);
			$dsn->original_recipient = $this->_canonicaliseAddress($matches[2]);
		}
		
		// Merak
		if (preg_match('/^Your message\s*To:\s*(\S+)\s*Subject:.+Sent:.+did not reach it\'s intended recipient. Please check the address and try again./ms',$message,$matches)) {
			$dsn->mta = self::MTA_MERAK;
			$dsn->type = self::TYPE_PERMANENT;
			$dsn->reason = self::REASON_RECIPIENT_UNKNOWN; // Fair assumption?
			$dsn->message = 'Your message did not reach it\'s intended recipient. Please check the address and try again.';
			$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
		}
		
		// SIMS - Sun Internet Mail Server
		if (strstr($message,'Subject: Delivery Notification: Delivery has failed')) {
			$dsn->mta = self::MTA_SIMS;
			$dsn->type = self::TYPE_PERMANENT;
			
			if (preg_match('/^\s*Reason: (.+)/m',$message,$matches)) {
				$dsn->message = $this->_cleanString($matches[1]);
			}
			
			// TODO: process To: header with RFC822
		}
		
		if (preg_match('/^From: PMDF e-Mail Interconnect/m',$message)) {
			$dsn->mta = self::MTA_PROCESS;
			
			// TODO: decode message
			if (preg_match('/^\s*Diagnostic code: smtp;(.+)/m',$message,$matches)) {
				$this->_processSMTPError($dsn,$matches[2]);
			}
		}
		
		
		// Process Software - probably PreciseMail Anti-Spam Gateway
		if ($dsn->mta == self::MTA_UNKNOWN && stristr($message, 'Error -- No such local user')) {
			$dsn->mta = self::MTA_PROCESS;
			$dsn->type = self::TYPE_PERMANENT;
			$dsn->reason = self::REASON_RECIPIENT_UNKNOWN;
			$dsn->message = 'Error -- No such local user';
			if (preg_match('/^\s+To: (.+)/m',$message,$matches)) {
				// FIXME: process RFC822
				$dsn->original_recipient = $this->_cleanString($matches[1]);
			}
		}
		
		// GroupWise
		if (preg_match('/^X-Mailer: Novell GroupWise/m',$message)) {
			$dsn->mta = self::MTA_GROUPWISE;
			if (preg_match('/^The message that you sent was opened by the following:\n(.+)\n/m',$message,$matches)) {
				$dsn->type = self::TYPE_READ;
				$dsn->original_recipient = $this->_canonicaliseAddress($matches[1]);
				$dsn->message = 'The message that you sent was opened';
			}
		}
		
		
		// Generic autoreply handling
		if ($dsn->type == self::TYPE_UNKNOWN) {
			// Could be a crap autoreply?
			if (preg_match('/will get back to you/i',$message)) {
				$dsn->type = self::TYPE_AUTOREPLY;
				// TODO: set $dsn->message to contain message body?
			}
			if (preg_match('/^Subject: Auto(:| Reply)/m',$message)) {
				$dsn->type = self::TYPE_AUTOREPLY;
				// TODO: set $dsn->message to contain message body?
			}
			if (preg_match('/^Subject: .+out of (the )? office/mi',$message)) {
				$dsn->type = self::TYPE_AUTOREPLY;
				// TODO: set $dsn->message to contain message body?
			}
			if (preg_match('|User-Agent: Vacation/|m',$message)) {
				$dsn->type = self::TYPE_AUTOREPLY;
				// TODO: set $dsn->message to contain message body?
			}
			// TODO Auto-Submitted: auto-replied
		}
		
		// General challenge handling
		if ($dsn->type == self::TYPE_UNKNOWN) {
			if (preg_match('/^Subject: (\[Challenge Response\]|Your message to .+ #\d+#$)/m',$message)) {
				$dsn->type = self::TYPE_CHALLENGE;
			}
		}
		
		// Final, generic processing
		if (!$dsn->smtp_code_enhanced && preg_match('/^Status:\s*(\d\.\d\.\d)( \((.+)\))?/m',$message,$matches)) {
			$dsn->smtp_code_enhanced = $matches[1];
			if (!$dsn->message && isset($matches[3])) {
				$dsn->message = $this->_cleanString($matches[3]);
			}
		}
		if (!$dsn->original_recipient && preg_match('/^(final|original)-recipient:\s*rfc822;(.+)/im',$message,$matches)) {$iv = str_repeat(chr(0),8);
			$dsn->original_recipient = $this->_canonicaliseAddress($matches[2]);
		}
		if ($dsn->type == self::TYPE_UNKNOWN && preg_match('/^action:\s*failed/im',$message)) {
			$dsn->type = self::TYPE_PERMANENT;
		}
		
		// See if we can determine a more specific reason
		$Generic = array(self::REASON_UNKNOWN,self::REASON_SMTP_MAIL,self::REASON_SMTP_RCPT,self::REASON_SMTP_DATA);
		if (in_array($dsn->reason,$Generic) && $dsn->message) {
			$dsn->reason = $this->_processReason($dsn->message,$dsn->reason);
		}
		
		// try again with whole message body
		// FIXME: ok for some but not for things like url detection => blacklisted
		/*	if (in_array($dsn->reason,$Generic)) {
			$dsn->reason = $this->_processReason($message,$dsn->reason);
		}*/
		
		// messages often have an SMTP enhanced code on the end
		if (preg_match('/\(#(\d\.\d\.\d)\)$/m',$dsn->message,$matches)) {
			$dsn->smtp_code_enhanced = $matches[1];
		}
		
		// Orig recipient fallback
		// TODO: more RFC compliant header parsing!
		if (!$dsn->original_recipient && $dsn->reason != self::REASON_NO_RECIPIENTS && preg_match('/^(From|Sender):.*([^<\s]+@[^>\s]+)>?\s*$/Umi',$message,$matches)) {
			$dsn->original_recipient = $matches[2];
		}

		// Tracking cookie check; this overrides all previous discovered forms of
		// original recipient
		if (preg_match('/' . self::$_tracking_header . ':\s*(\S+)/m', $message, $matches)) {
			$dsn->Debug = "Found tracking token $matches[1]";
			$recipient = $this->decodeTrackingToken($matches[1]);
			if ($recipient !== false) {
				$dsn->original_recipient = $recipient;
			}
		}	
		
		return $dsn;
	}
	
	protected function _cleanString($String)
	{
		return trim(preg_replace('/[\r\n]\s*/',' ',$String));
	}
	
	protected function _canonicaliseAddress($Address)
	{
		// TODO: do this properly according to RFC822
		return trim(preg_replace('/[<>]/','',$Address));
	}
	
	#Destinatario desconocido o mailbox lleno = unknown addressee or mailbox full
	protected function _processReason($message,$ExistingReason=self::REASON_UNKNOWN)
	{
		if (preg_match('/user unknown|unknown user|unknown local.?part|mailbox not found|unrouteable address|no mailbox here by that name|recipient name is not recognized|not a valid (email|user|mailbox)|doesn\'t have a (btinternet\.com|btopenworld\.com|yahoo\..{2,6}) account|unsupported mail destination|invalid recipient|user does not exist|unable to validate recipient/i',$message)) {
			return self::REASON_RECIPIENT_UNKNOWN;
		}
		
		if (preg_match('/quota exceeded|exceeded quota|mailbox (is )?full|over.?quota|reached disk quota/i',$message)) {
			return self::REASON_OVERQUOTA;
		}
		
		if (preg_match('/sender address \S+ does not resolve/i',$message)) {
			return self::REASON_SENDER_DNS;
		}
		
		if (preg_match('/is not accepting mail from this sender/i',$message)) {
			return self::REASON_RECIPIENT_BLOCKED_SENDER;
		}
		
		if (preg_match('/relay(ing)? (denied|blocked)|relaying to.+prohibited/i',$message)) {
			return self::REASON_RELAY_DENIED;
		}
		
		if (preg_match('/account (has been |is )?disabled|inactive user/i',$message)) {
			return self::REASON_RECIPIENT_DISABLED;
		}
		
		if (preg_match('/file (.+ )?has been blocked/i',$message)) {
			return self::REASON_CONTENT;
		}
		
		if (preg_match('/'.implode('|',$this->_blacklists_ip).'/i',$message)) {
			return self::REASON_BLACKLISTED_HOST;
		}
		
		if (preg_match('/'.implode('|',$this->_blacklists_domain).'/i',$message)) {
			return self::REASON_BLACKLISTED_HOST;
		}
		
		if (stristr($message,'http:') || stristr($message,'blacklisted')) {
			return self::REASON_BLACKLISTED;
		}
		
		return $ExistingReason;
	}
	
	protected function _processSMTPError(&$dsn, $message)
	{
		$message = $this->_cleanString($message);

		if (preg_match('/^((\d\d\d)[- ])?(\d\.\d\.\d )?(.+)/',$message,$matches)) {
		
			$message = preg_replace('/\d{3}[- ]/','',$message);
			if (!empty($matches[2])) $dsn->smtp_code = trim($matches[2]);
			if (!empty($matches[3])) $dsn->smtp_code_enhanced = trim($matches[3]);
			
			// Strip SMTP continuation markers and save message
			if (!empty($matches[4]) && empty($dsn->message)) {
				$dsn->message = preg_replace('/\s\d{3}[- ]/',' ',$matches[4]);
			}
			
			if ($dsn->smtp_code && $dsn->type == self::TYPE_UNKNOWN) {
				if (substr($dsn->smtp_code,0,1) == '5') {
					$dsn->type = self::TYPE_PERMANENT;
				} else if (substr($dsn->smtp_code,0,1) == '4') {
					$dsn->type = self::TYPE_TEMPORARY;
				}
			}
		}
	}
	
	public static function getTrackingHeaderName()
	{
		return self::$_tracking_header;
	}
	
	public static function getTrackingToken($email, $key)
	{
		if (empty($email)) {
			throw new Exception('Email e-mail address passed');
		}
		if (empty($key)) {
			throw new Exception('Empty key passed');
		}
		self::checkRuntimeDependencies();
		$iv = str_repeat(chr(0), mcrypt_get_iv_size(MCRYPT_BLOWFISH, MCRYPT_MODE_CFB)); // Not very secure, but we don't need it to be!
		$base64 = base64_encode(mcrypt_encrypt(MCRYPT_BLOWFISH, $key, $email, MCRYPT_MODE_CFB, $iv));
		// To make it even safer (e.g. for use in e-mail VERP schemes) we
		// use a variant of base64 URL encoding to remove "+", "/" and "=" characters
		// We also wrap it with an "A" at the start and end
		return 'A' . strtr($base64, '+/=', '-_.') . 'A';
	}
	
	public function decodeTrackingToken($token, $key)
	{
		$this->checkRuntimeDependencies();
		$iv = str_repeat(chr(0), mcrypt_get_iv_size(MCRYPT_BLOWFISH, MCRYPT_MODE_CFB));
		$token = substr($token, 1, -1); //strip 'A' wrappers
		return mcrypt_decrypt(MCRYPT_BLOWFISH, $key, base64_decode(strtr($token, '-_.', '+/=')), MCRYPT_MODE_CFB, $iv);
	}
	
	public static function checkRuntimeDependencies()
	{
		if (!function_exists('mcrypt_encrypt')) {
			throw new Exception("You must have the 'mcrypt' extension installed");
		}
		
		if (!in_array(MCRYPT_BLOWFISH, mcrypt_list_algorithms())) {
			throw new Exception('Your version of PHP does not support Blowfish encryption');
		}
		
		return true;
	}
}

